/**
 * Copyright (C) 2014 EDIT
 * European Distributed Institute of Taxonomy
 * http://www.e-taxonomy.eu
 *
 * The contents of this file are subject to the Mozilla Public License Version 1.1
 * See LICENSE.TXT at the top of this package for the full license terms.
 */
package eu.etaxonomy.cdm.api.application;

import java.util.List;

import org.apache.logging.log4j.LogManager;import org.apache.logging.log4j.Logger;
import org.hibernate.collection.internal.AbstractPersistentCollection;
import org.hibernate.proxy.AbstractLazyInitializer;
import org.springframework.context.ApplicationListener;
import org.springframework.core.io.ClassPathResource;
import org.springframework.core.io.Resource;

import eu.etaxonomy.cdm.api.service.ITestService;
import eu.etaxonomy.cdm.cache.CdmRemoteCacheManager;
import eu.etaxonomy.cdm.common.monitor.IProgressMonitor;
import eu.etaxonomy.cdm.common.monitor.NullProgressMonitor;
import eu.etaxonomy.cdm.ext.geo.IEditGeoService;
import eu.etaxonomy.cdm.io.service.IIOService;
import eu.etaxonomy.taxeditor.remoting.source.ICdmRemoteSource;
import eu.etaxonomy.taxeditor.service.ICachedCommonService;
import eu.etaxonomy.taxeditor.session.ICdmEntitySessionManager;

/**
 * CDM Application Controller class for remoting clients
 *
 * FIXME:Remoting extending {@link CdmApplicationController} is a temporary workaround.
 * The {@link CdmApplicationController} should be split into a CdmApplicationControllerBase
 * class with {@link CdmApplicationController} and this class as subclasses
 */
public class CdmApplicationRemoteController  extends CdmApplicationController {

    private static final Logger logger = LogManager.getLogger(CdmApplicationRemoteController.class);

    public static final Resource DEFAULT_REMOTE_APPLICATION_CONTEXT_RESOURCE =
            new ClassPathResource("/eu/etaxonomy/cdm/remotingApplicationContext.xml");
    private final Resource applicationContextResource;
    private final IProgressMonitor progressMonitor;
    private final ICdmRemoteSource remoteSource;

//    public static CdmApplicationRemoteController NewInstance(Resource applicationContextResource,
//            ICdmRemoteSource remoteSource,
//            IProgressMonitor progressMonitor,
//            List<ApplicationListener> listeners) {
//        return new CdmApplicationRemoteController(applicationContextResource,
//                remoteSource,
//                false,
//                progressMonitor,
//                listeners);
//
//    }

    /**
     * Creates new instance of CdmApplicationRemoteController
     *
     * @param applicationContextResource
     * @param remoteSource
     * @param omitTermLoading
     * @param progressMonitor
     * @param listeners
     * @return
     */
    public static CdmApplicationRemoteController NewInstance(ICdmRemoteSource remoteSource,
            IProgressMonitor progressMonitor,
            List<ApplicationListener> listeners) {

        return new CdmApplicationRemoteController(DEFAULT_REMOTE_APPLICATION_CONTEXT_RESOURCE,
                remoteSource,
                false,
                progressMonitor,
                listeners);
    }

    public static CdmApplicationRemoteController NewInstance(ICdmRemoteSource remoteSource,
            boolean validateXml,
            IProgressMonitor progressMonitor,
            List<ApplicationListener> listeners) {

        return new CdmApplicationRemoteController(DEFAULT_REMOTE_APPLICATION_CONTEXT_RESOURCE,
                remoteSource,
                validateXml,
                progressMonitor,
                listeners);
    }

    /**
     * Constructs CdmApplicationRemoteController
     *
     * @param applicationContextResource
     * @param remoteSource
     * @param omitTermLoading
     * @param progressMonitor
     * @param listeners
     */
    private CdmApplicationRemoteController(Resource applicationContextResource,
            ICdmRemoteSource remoteSource,
            boolean validateXml,
            IProgressMonitor progressMonitor,
            List<ApplicationListener> listeners){
        logger.info("Start CdmApplicationRemoteController with remote source: " + remoteSource.getName());
        this.applicationContextResource =
                applicationContextResource != null ? applicationContextResource : DEFAULT_REMOTE_APPLICATION_CONTEXT_RESOURCE;
        this.progressMonitor = progressMonitor != null ? progressMonitor : new NullProgressMonitor();
        this.remoteSource = remoteSource;
        setNewRemoteSource(remoteSource, validateXml, listeners);

    }

    /**
     * Creates and starts a new spring application context
     *
     * @param remoteSource object for connecting to an http-invoker server
     * @param omitTermLoading
     * @param listeners
     * @return
     */
    public boolean setNewRemoteSource(ICdmRemoteSource remoteSource,
            boolean validateXml,
            List<ApplicationListener> listeners){

        logger.info("Connecting to '" + remoteSource.getName() + "'");

        int nTasks = 2;

        progressMonitor.beginTask("Connecting to '" + remoteSource.getName() + "'", nTasks);

        progressMonitor.subTask("Initialising CDM Model Cache ...");
        CdmRemoteCacheManager.getInstance();
        CdmRemoteCacheManager.removeEntityCaches();
        progressMonitor.worked(1);
        progressMonitor.subTask("Starting application context ...");
        init();
        progressMonitor.worked(1);
        progressMonitor.done();
        return true;
    }

    @Override
    protected void init(){
        configuration = new CdmApplicationRemoteConfiguration(remoteSource);
        ((CdmApplicationRemoteConfiguration)configuration).setApplicationContext(applicationContext);
        AbstractLazyInitializer.setConfiguration((CdmApplicationRemoteConfiguration)configuration);
        AbstractPersistentCollection.setConfiguration((CdmApplicationRemoteConfiguration)configuration);
    }

    public ICdmEntitySessionManager getCdmEntitySessionManager() {
        return ((CdmApplicationRemoteConfiguration)configuration).getCdmEntitySessionManager();
    }

    public IIOService getIOService(){
        return ((CdmApplicationRemoteConfiguration)configuration).getIOService();
    }

    public ITestService getTestService(){
        return ((CdmApplicationRemoteConfiguration)configuration).getTestService();
    }

    public ICachedCommonService getCachedCommonService(){
        return ((CdmApplicationRemoteConfiguration)configuration).getCachedCommonService();
    }

    public IEditGeoService getEditGeoService() {
        return ((CdmApplicationRemoteConfiguration)configuration).getEditGeoService();
    }
}
