/**
 * Copyright (C) 2015 EDIT
 * European Distributed Institute of Taxonomy
 * http://www.e-taxonomy.eu
 *
 * The contents of this file are subject to the Mozilla Public License Version 1.1
 * See LICENSE.TXT at the top of this package for the full license terms.
 */
package eu.etaxonomy.taxeditor.remoting.source;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.net.InetSocketAddress;
import java.net.Socket;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

import org.apache.commons.lang3.StringUtils;
import org.apache.http.HttpEntity;
import org.apache.http.HttpResponse;
import org.apache.http.client.ClientProtocolException;
import org.apache.http.client.ResponseHandler;
import org.apache.http.client.config.RequestConfig;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.impl.client.CloseableHttpClient;
import org.apache.http.impl.client.HttpClientBuilder;
import org.apache.http.util.EntityUtils;
import org.apache.logging.log4j.LogManager;import org.apache.logging.log4j.Logger;
import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;

import com.fasterxml.jackson.core.type.TypeReference;
import com.fasterxml.jackson.databind.ObjectMapper;

import eu.etaxonomy.cdm.api.application.CdmApplicationRemoteConfiguration;
import eu.etaxonomy.cdm.api.application.CdmApplicationState;
import eu.etaxonomy.cdm.common.CdmUtils;
import eu.etaxonomy.cdm.config.CdmSourceException;
import eu.etaxonomy.cdm.config.ConfigFileUtil;
import eu.etaxonomy.cdm.config.ICdmSource;
import eu.etaxonomy.cdm.database.CdmPersistentDataSource;
import eu.etaxonomy.cdm.database.ICdmDataSource;
import eu.etaxonomy.cdm.model.metadata.CdmMetaData;
import eu.etaxonomy.taxeditor.remoting.server.CdmServerException;
import eu.etaxonomy.taxeditor.remoting.server.CdmServerUtils;

/**
 * @author cmathew
 * @date 20 Jan 2015
 */
public class CdmServerInfo {

    public static final Logger logger = LogManager.getLogger(CdmServerInfo.class);

    protected final static String CDMSERVER_PREFIX = "cdmserver/";
    private final static String NAME_PRODUCTION = "cybertaxonomy.org";
    private final static String SERVER_PRODUCTION = "api.cybertaxonomy.org";

    private final static String NAME_DEMO_1 = "demo I";
    private final static String SERVER_DEMO_1 = "160.45.63.230";

    public final static String SERVER_LOCALHOST = "localhost";
    private final static String NAME_LOCALHOST = "localhost";
    public final static String NAME_LOCALHOST_MGD = "localhost mgd.";

    private final static String NAME_LOCALHOST_DEV = "localhost-dev";
    private final static String NAME_INSTANCE_LOCALHOST_DEV = "local-dev";
    private final static String SERVER_LOCALHOST_DEV = "localhost";
    private final static String BASEPATH_LOCALHOST_DEV = "";

    public final static int NULL_PORT = -1;
    public final static String NULL_PORT_STRING = "N/A";

    protected static final String CDM_REMOTE_SERVERS_CONFIG_FILE = "cdm_remote_servers.json";


    private final String name;
    private final String server;
    private int port;
    protected final List<CdmInstanceInfo> instances;

    private String cdmlibServicesVersion = "";
    private String cdmlibServicesLastModified = "";

    private String prefix = "";

    private boolean ignoreCdmLibVersion = false;


    public CdmServerInfo(CdmServerInfoConfig parameterObject) {
        this.name = parameterObject.getName();
        this.server = parameterObject.getServer();
        this.port = parameterObject.getPort();
        this.prefix = parameterObject.getPrefix();
        this.ignoreCdmLibVersion = parameterObject.isIgnoreCdmLibVersion();
        this.instances = new ArrayList<>();
    }

    public CdmInstanceInfo addInstance(String name, ICdmDataSource dataSource) {
        CdmInstanceInfo cii = new CdmInstanceInfo(name, dataSource);
        instances.add(cii);
        return cii;
    }

    public CdmInstanceInfo addInstance(String name, String basePath) {
        CdmInstanceInfo cii = new CdmInstanceInfo(name, basePath);
        instances.add(cii);
        return cii;
    }

    public boolean isLocalhost() {
        return name.startsWith(SERVER_LOCALHOST);
    }

    public boolean isLocalhostMgd() {
        return NAME_LOCALHOST_MGD.equals(name);
    }

    public String getCdmlibServicesVersion() {
        return cdmlibServicesVersion;
    }

    public String getCdmlibLastModified() {
        return cdmlibServicesLastModified;
    }

    public void refreshInstances() throws CdmServerException {
        instances.clear();
        if(isLocalhostMgd()) {
            addInstancesFromDataSourcesConfig();
        } else {
            addInstancesViaHttp();
        }


        Collections.sort(instances, new Comparator<CdmInstanceInfo>() {
            @Override
            public int compare(CdmInstanceInfo cii1, CdmInstanceInfo cii2){
                return cii1.getName().toString().compareTo(cii2.getName().toString());
            }
        });
    }

    public void updateInfo() throws CdmServerException {

        String url = guessProtocol() + "://" + server + ":" + String.valueOf(port) + "/" + prefix + "info.jsp";
        String responseBody = getResponse(url);
        if(responseBody != null) {
            try {
                JSONObject info = new JSONObject(responseBody);
                cdmlibServicesVersion =  info.getString("cdmlibServicesVersion");
                cdmlibServicesLastModified = info.getString("cdmlibServicesLastModified");
            } catch (JSONException e) {
                throw new CdmServerException(e);
            }
        }
    }

    String guessProtocol() {
        return port == 443 ? "https" : "http";
    }

    public void addInstancesViaHttp() throws CdmServerException {
        updateInfo();
        String url = guessProtocol() + "://" + server + ":" + String.valueOf(port) + "/" + prefix + "instances.jsp";
        String responseBody = getResponse(url);
        if(responseBody != null) {
            try {
                JSONArray array = new JSONArray(responseBody);
                for(int i=0;i<array.length();i++) {
                    JSONObject instance = (JSONObject)array.get(i);
                    if(instance != null) {
                        JSONObject conf = (JSONObject)instance.get("configuration");
                        if(conf != null) {
                            String instanceName = conf.getString("instanceName");
                            // we need to remove the first (char) forward slash from
                            // the base path
                            String basePath = conf.getString("basePath").substring(1);
                            addInstance(instanceName, basePath);
                            logger.info("Added instance with name : " + instanceName + ", basePath : " + basePath);
                        }
                    }
                }
            } catch (JSONException e) {
                throw new CdmServerException(e);
            }
        }
    }

    private String getResponse(String url) throws CdmServerException {

        RequestConfig.Builder requestBuilder = RequestConfig.custom();
        requestBuilder.setConnectTimeout(CdmApplicationRemoteConfiguration.HTTP_READ_TIMEOUT_MIN);
        requestBuilder.setConnectionRequestTimeout(CdmApplicationRemoteConfiguration.HTTP_READ_TIMEOUT_MIN);
        requestBuilder.setSocketTimeout(CdmApplicationRemoteConfiguration.HTTP_READ_TIMEOUT_MIN);
        CloseableHttpClient client = HttpClientBuilder.create().setDefaultRequestConfig(requestBuilder.build()).build();

        HttpGet httpGet = new HttpGet(url);

        logger.info("Executing request " + httpGet.getRequestLine());

        // Create a custom response handler
        ResponseHandler<String> responseHandler = new ResponseHandler<String>() {

            @Override
            public String handleResponse(
                    final HttpResponse response) throws ClientProtocolException, IOException {
                int status = response.getStatusLine().getStatusCode();
                if (status >= 200 && status < 300) {
                    HttpEntity entity = response.getEntity();
                    return entity != null ? EntityUtils.toString(entity) : null;
                } else {
                    throw new ClientProtocolException("Unexpected response status: " + status);
                }
            }

        };
        String responseBody = null;
        try {
            responseBody = client.execute(httpGet, responseHandler);
        } catch (ClientProtocolException e) {
            throw new CdmServerException(e);
        } catch (IOException e) { // java.net.ConnectException: Connection refused, java.net.SocketTimeoutException: Read timed out
            throw new CdmServerException(e);
        } finally{
            try {
                client.close();
                client = null;
            } catch (IOException e) {
                // IGNORE //
            }
        }
        return responseBody;
    }

    public void addInstancesFromDataSourcesConfig() {
        for(ICdmDataSource dataSource : CdmPersistentDataSource.getAllDataSources()){
            String datasourceNCName = CdmServerUtils.xmlNCNamefrom(dataSource.getName());
            logger.info("Adding local instance " + dataSource.getName() + " as " + datasourceNCName);
            addInstance(datasourceNCName, dataSource);
        }
    }

    public String toString(String instanceName, int port) {
        return server + ":" + String.valueOf(port) + "/" + instanceName;
    }

    public CdmInstanceInfo getInstanceFromName(String instanceName) {
        if(instanceName == null) {
            return null;
        }

        for(CdmInstanceInfo instance : instances) {
            if(instance.getName() != null && instance.getName().equals(instanceName)) {
                return instance;
            }
        }
        return null;
    }

    public ICdmRemoteSource getCdmRemoteSource(CdmInstanceInfo instance, int port) {
        if(instance != null) {
            if (instance.getDataSource()!=null){
                return CdmRemoteLocalhostSource.NewInstance(name,
                        server,
                        port,
                        instance.getDataSource(),
                        instance.name
                        );
            }else{
                return CdmRemoteSource.NewInstance(name,
                        server,
                        port,
                        instance.getBasePath()
                        );
            }
        }
        return null;
    }

    public boolean pingServer() throws CdmServerException, IOException {
        if(isLocalhostMgd()) {
            return true;
        }
        Socket s = new Socket();
        s.connect(new InetSocketAddress(server, port), CdmApplicationRemoteConfiguration.HTTP_READ_TIMEOUT_MIN);
        s.close();
        logger.info("[CDM-Server] Available @ " + server + ":" + port );
        updateInfo();
        return true;
    }

    public boolean pingInstance(CdmInstanceInfo instance, int port) throws CdmServerException  {
        ICdmSource cdmRemoteSource = getCdmRemoteSource(instance, port);
        try {
            if(cdmRemoteSource != null && cdmRemoteSource.checkConnection()) {
                logger.info("[CDM-Server] Running @ " + server + ":" + port + " for instance " + instance);
                return true;
            }
        } catch (CdmSourceException e) {
            logger.error(e.getMessage(), e);
            throw new CdmServerException(e);
        }

        return false;
    }

    public int compareDbSchemaVersion(CdmInstanceInfo instance, int port) throws CdmServerException {

        ICdmSource remoteSource = getCdmRemoteSource(instance, port);
        String remoteDbSchemaVersion;
        try {
            remoteDbSchemaVersion = remoteSource.getDbSchemaVersion();
        } catch (CdmSourceException e) {
            throw new CdmServerException(e);
        }

        if(remoteDbSchemaVersion != null) {
            return CdmMetaData.compareVersion(remoteDbSchemaVersion, CdmMetaData.getDbSchemaVersion(), 3, null);
        } else {
            throw new CdmServerException("Cannot determine editor db schema version");
        }
    }

    public int compareCdmlibServicesVersion() throws CdmSourceException {

        String serverVersion = cdmlibServicesVersion;
        String serverCdmlibLastModified = cdmlibServicesLastModified;
        if(ignoreCdmLibVersion) {
            return 0;
        } else {
            return compareCdmlibServicesVersion(serverVersion, serverCdmlibLastModified);
        }
    }

    public static int compareCdmlibServicesVersion(String serverVersion, String serverCdmlibLastModified) throws CdmSourceException {

        String editorVersion = CdmApplicationState.getCdmlibVersion();
        String editorCdmlibLastModified = CdmApplicationState.getCdmlibLastModified();

        int result = 0;
        if(StringUtils.isBlank(serverVersion) || StringUtils.isBlank(editorVersion)) {
            throw new CdmSourceException("cdmlib-services server or editor version is empty");
        }

        String[] serverVersionSplit = serverVersion.split("\\.");
        String[] editorVersionSplit = editorVersion.split("\\.");

        if(serverVersionSplit.length < 3 || editorVersionSplit.length < 3 || serverVersionSplit.length > 4 || editorVersionSplit.length > 4) {
            throw new CdmSourceException("cdmlib-services server or editor version is invalid");
        }

        Integer serverVersionPart;
        Integer editorVersionPart;

        for(int i=0 ; i<3 ; i++) {
            serverVersionPart = Integer.valueOf(serverVersionSplit[i]);
            editorVersionPart = Integer.valueOf(editorVersionSplit[i]);

            int partCompare = serverVersionPart.compareTo(editorVersionPart);
            if (partCompare != 0){
                return partCompare;
            }
        }
        // at this point major, minor and patch versions are matching

        if(StringUtils.isBlank(serverCdmlibLastModified) || StringUtils.isBlank(editorCdmlibLastModified)) {
            throw new CdmSourceException("cdmlib-services server or editor version is empty");
        }

        String cdmServerIgnoreVersion = System.getProperty("cdm.server.version.lm.ignore");
        if(StringUtils.isBlank(cdmServerIgnoreVersion) || !cdmServerIgnoreVersion.equals("true")) {
            Long serverLastModified = Long.valueOf(serverCdmlibLastModified);
            Long editorLastModified = Long.valueOf(editorCdmlibLastModified);
            return serverLastModified.compareTo(editorLastModified);
        }

        return result;
    }

    public static List<CdmServerInfo> getCdmServers(boolean isLocal) {
        List<CdmServerInfoConfig> configList;
        File file = new File(ConfigFileUtil.perUserCdmFolderFallback(), CDM_REMOTE_SERVERS_CONFIG_FILE);
        if (file.exists()){
            configList = loadFromConfigFile(file);
        }else{
            configList = loadFromConfigFile(new File(ConfigFileUtil.perUserCdmFolderFallback(), CDM_REMOTE_SERVERS_CONFIG_FILE));
        }
        List<CdmServerInfo> serverInfoList = new ArrayList<>(configList.size());
        for(CdmServerInfoConfig config : configList) {
            serverInfoList.add(new CdmServerInfo(config));
        }
        // The local host managed server must not be in the config file, this should be moved and only added when plugin is installed
        if (isLocal){
            CdmServerInfoConfig localHostManagedConfig = new CdmServerInfoConfig(NAME_LOCALHOST_MGD, SERVER_LOCALHOST, NULL_PORT, CDMSERVER_PREFIX, false);
            serverInfoList.add(new CdmServerInfo(localHostManagedConfig));
        }

        return serverInfoList;
    }

    protected static List<CdmServerInfoConfig>  loadFromConfigFile(File file) {

        List<CdmServerInfoConfig> serverList = null;

        ObjectMapper mapper = new ObjectMapper();

        if(!file.exists()) {
            logger.info("The remote server config file '" + file.getAbsolutePath() +
                    "' does not yet exist, it will be created based on the defaults.");
             try {
                serverList = createDefaultServerConfigList();
                mapper.writerWithDefaultPrettyPrinter().writeValue(new FileOutputStream(file), serverList);

            } catch (IOException e) {
                throw new RuntimeException(e);
            }
        } else {
            try {
                serverList = mapper.readValue(new FileInputStream(file), new TypeReference<List<CdmServerInfoConfig>>(){});
            } catch (IOException e) {
               throw new RuntimeException(e);
            }
        }

        return serverList;
    }

    private static List<CdmServerInfoConfig> createDefaultServerConfigList() {

        List<CdmServerInfoConfig> serverInfoList = new ArrayList<>();
        serverInfoList.add(new CdmServerInfoConfig(NAME_PRODUCTION, SERVER_PRODUCTION, 443, "", false));
        serverInfoList.add(new CdmServerInfoConfig(NAME_DEMO_1, SERVER_DEMO_1, 80, CDMSERVER_PREFIX, false));
        serverInfoList.add(new CdmServerInfoConfig(NAME_LOCALHOST, SERVER_LOCALHOST, 8080, CDMSERVER_PREFIX, true));
        return serverInfoList;
    }

    public String getName() {
        return name;
    }

    public String getServer() {
        return server;
    }

    public int getPort() {
        return port;
    }

    public void setPort(int port) {
        this.port = port;
    }

    public List<CdmInstanceInfo> getInstances() throws CdmServerException {
        if(instances.isEmpty()) {
            refreshInstances();
        }
        return instances;
    }

    public static ICdmSource getDevServerRemoteSource() {
        String value = System.getProperty("cdm.server.dev.port");
        boolean available = false;
        CdmInstanceInfo devInstance = null;
        if(CdmUtils.isNotBlank(value)) {
            int devPort = Integer.valueOf(value);
            CdmServerInfo devCii = new CdmServerInfo(new CdmServerInfoConfig(NAME_LOCALHOST_DEV, SERVER_LOCALHOST_DEV, devPort, "", false));
            try {
                devInstance = devCii.addInstance(NAME_INSTANCE_LOCALHOST_DEV, BASEPATH_LOCALHOST_DEV);
                available = devCii.pingInstance(devInstance, devPort);
                if(available) {
                    logger.info("Will connect local development cdm instance: at port " + devPort);
                    return devCii.getCdmRemoteSource(devInstance, devPort);
                }
            } catch (CdmServerException e) {
                logger.error("Can not connect to local development cdm instance at port " + devPort + ". "
                        + "Make sure the cdm instance is running and that the Spring profile \"remoting\" is "
                        + "activated (-Dspring.profiles.active=remoting)", e);
                //TODO show message dialog only instead of throwing the exception to show the error
                // dialog is not necessary in this situation
            }
            logger.error("local development cdm instance at port " + devPort + " is not accessible");
        }
        return null;
    }

    public class CdmInstanceInfo {

        private final String name;

        /**
         * The full path of the instance including the the prefix (if any).
         * E.g. for an EDIT instance this would be something like "cdmserver/remoting"
         * For a managed local server this would simply be ""
         */
        private final String basePath;

        private final ICdmDataSource dataSource;

        public CdmInstanceInfo(String name, String basePath) {
            this.name = name;
            this.basePath = basePath;
            this.dataSource = null;
        }

        public CdmInstanceInfo(String name, ICdmDataSource dataSource) {
            this.name = name;
            this.basePath = "";
            this.dataSource = dataSource;
        }

        public String getName() {
            return name;
        }

        public String getBasePath() {
            return basePath;
        }

        public ICdmDataSource getDataSource() {
            return dataSource;
        }
    }
}
