/**
* Copyright (C) 2020 EDIT
* European Distributed Institute of Taxonomy
* http://www.e-taxonomy.eu
*
* The contents of this file are subject to the Mozilla Public License Version 1.1
* See LICENSE.TXT at the top of this package for the full license terms.
*/
package eu.etaxonomy.taxeditor.service;

import java.io.IOException;
import java.net.URI;
import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;

import org.apache.http.Header;
import org.apache.http.HttpResponse;
import org.apache.http.client.HttpClient;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.config.Registry;
import org.apache.http.config.RegistryBuilder;
import org.apache.http.conn.socket.ConnectionSocketFactory;
import org.apache.http.conn.socket.PlainConnectionSocketFactory;
import org.apache.http.conn.ssl.SSLConnectionSocketFactory;
import org.apache.http.impl.NoConnectionReuseStrategy;
import org.apache.http.impl.client.HttpClientBuilder;
import org.apache.http.impl.conn.PoolingHttpClientConnectionManager;
import org.springframework.remoting.httpinvoker.HttpComponentsHttpInvokerRequestExecutor;
import org.springframework.remoting.httpinvoker.HttpInvokerClientConfiguration;

import eu.etaxonomy.taxeditor.remoting.RemoteExecutionTimestampsUtil;

/**
 * HttpInvokerRequestExecutor which extends the spring
 * {@link HttpComponentsHttpInvokerRequestExecutor} by two functionalities:
 *
 * <ol>
 * <li>Records time stamps when sending the request and when receiving the response.</li>
 * <li>Allows custom configuration of the {@link PoolingHttpClientConnectionManager}
 * (see <a href="https://dev.e-taxonomy.eu/redmine/issues/8812">https://dev.e-taxonomy.eu/redmine/issues/8812</a>).
 * </ol>
 *
 * @author a.kohlbecker
 * @since Jan 17, 2020
 */
public class TimestampingHttpInvokerRequestExecutor extends HttpComponentsHttpInvokerRequestExecutor {

    private static final int DEFAULT_MAX_TOTAL_CONNECTIONS = 100;

    private static final int DEFAULT_MAX_CONNECTIONS_PER_ROUTE = 5;

    /**
     * Default in PoolingHttpClientConnectionManager is 2000 milliseconds.
     *
     * see https://stackoverflow.com/questions/10558791/apache-httpclient-interim-error-nohttpresponseexception
     */
    private static final int VALIDATE_AFTER_INACTIVITY = 2000;

    public TimestampingHttpInvokerRequestExecutor() {
        super(createDefaultHttpClient());
    }

    @Override
    protected HttpResponse executeHttpPost(HttpInvokerClientConfiguration config, HttpClient httpClient,
            HttpPost httpPost) throws IOException {

        setExecutionRequestClientTimestamp(httpPost.getURI());

        HttpResponse respone = super.executeHttpPost(config, httpClient, httpPost);

        setExecutionResponseHttpHeaderTimestamp(respone);

        return respone;
    }

    private void setExecutionRequestClientTimestamp(URI requestURI){

        LocalDateTime date = LocalDateTime.now();
        String dateTimeStr = date.format(DateTimeFormatter.ISO_DATE_TIME);
        RemoteExecutionTimestampsUtil.setLastRequestClientTime(dateTimeStr);
        if(requestURI != null){
            RemoteExecutionTimestampsUtil.setLastServiceMethod(requestURI.toString());
        }
    }

    private void setExecutionResponseHttpHeaderTimestamp(HttpResponse respone){
        Header dateHeader = respone.getFirstHeader("Date");
        if(dateHeader != null){
            RemoteExecutionTimestampsUtil.setLastResponseHttpHeaderTime(dateHeader.getValue());
        }
    }

    private static HttpClient createDefaultHttpClient() {
        Registry<ConnectionSocketFactory> schemeRegistry = RegistryBuilder.<ConnectionSocketFactory>create()
                .register("http", PlainConnectionSocketFactory.getSocketFactory())
                .register("https", SSLConnectionSocketFactory.getSocketFactory())
                .build();

        PoolingHttpClientConnectionManager connectionManager = new PoolingHttpClientConnectionManager(schemeRegistry);
        connectionManager.setMaxTotal(DEFAULT_MAX_TOTAL_CONNECTIONS);
        connectionManager.setDefaultMaxPerRoute(DEFAULT_MAX_CONNECTIONS_PER_ROUTE);
        connectionManager.setValidateAfterInactivity(VALIDATE_AFTER_INACTIVITY);

        return HttpClientBuilder.create()
                .setConnectionManager(connectionManager)
                .setConnectionReuseStrategy(new NoConnectionReuseStrategy()) // see #8812
                .build();
    }
}