/**
 * Copyright (C) 2014 EDIT
 * European Distributed Institute of Taxonomy
 * http://www.e-taxonomy.eu
 *
 * The contents of this file are subject to the Mozilla Public License Version 1.1
 * See LICENSE.TXT at the top of this package for the full license terms.
 */
package eu.etaxonomy.taxeditor.session;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.logging.log4j.LogManager;import org.apache.logging.log4j.Logger;
import org.springframework.stereotype.Component;

import eu.etaxonomy.cdm.api.service.UpdateResult;
import eu.etaxonomy.cdm.model.common.CdmBase;
import eu.etaxonomy.cdm.model.common.ICdmBase;
import eu.etaxonomy.cdm.persistence.dto.MergeResult;

/**
 * Implementation for {@link ICdmEntitySessionManager}
 *
 * @author cmathew
 * @date 16 Oct 2014
 */
@Component
public class CdmEntitySessionManager implements ICdmEntitySessionManager {

    private static final Logger logger = LogManager.getLogger(CdmEntitySessionManager.class);

    private final Map<ICdmEntitySessionEnabled, ICdmEntitySession> ownerSessionMap = new HashMap<>();

    private final List<ICdmEntitySessionManagerObserver> sessionObservers = new ArrayList<>();

    //private ICdmEntitySession activeSession;

    private final InheritableThreadLocal<ICdmEntitySession> tlActiveSession = new InheritableThreadLocal<>();

    private NullSession nullSession;

    @Override
    public ICdmEntitySession getNullSession() {
        return nullSession;
    }

    @Override
    public ICdmEntitySession newSession(ICdmEntitySessionEnabled sessionOwner, boolean setAsActive) {
        ICdmEntitySession session = ownerSessionMap.get(sessionOwner);
        if(session == null) {
            session = new CdmEntitySession(sessionOwner, this);
            addToOwnerSessionMap(sessionOwner, session);
        }

        if(setAsActive) {
            setActiveSession(session);
        }

        return session;
    }

    @Override
    public ICdmEntitySession bindNullSession() {

        if(nullSession == null) {
            nullSession = new NullSession(null, this);
        }

        setActiveSession(nullSession);

        return nullSession;
    }

    @Override
    public ICdmEntitySession getActiveSession() {
        return tlActiveSession.get();
    }

    private void setActiveSession(ICdmEntitySession activeSession) {
       this. tlActiveSession.set(activeSession);
        notifyObservers();
    }

    @Override
    public Collection<ICdmEntitySession> getSessions() {
        return ownerSessionMap.values();
    }

    @Override
    public void bind(ICdmEntitySessionEnabled sessionOwner) {
        if(sessionOwner == null) {
            setActiveSession(null);
            return;
        }
        ICdmEntitySession session  = ownerSessionMap.get(sessionOwner);
        if(session == null) {
            throw new CdmClientSessionException("Trying to bind session which does not exist");
        }

        setActiveSession(session);
    }

    @Override
    public boolean contains(ICdmEntitySessionEnabled sessionOwner) {
        return ownerSessionMap.containsKey(sessionOwner);
    }

    @Override
    public <T extends Object> T load(T obj, boolean update) {
        if(tlActiveSession.get() == null) {
            return obj;
        } else {
            return tlActiveSession.get().load(obj, update);
        }
    }

    @Override
    public <T extends CdmBase> void update() {
        if(tlActiveSession.get() != null) {
            tlActiveSession.get().update();
        }
    }

    @Override
    public <T extends CdmBase> T load(T cdmBase, boolean update) {
        if(tlActiveSession.get() == null) {
            return cdmBase;
        }
        return tlActiveSession.get().load(cdmBase, update);
    }

    @Override
    public UpdateResult load(UpdateResult updateResult, boolean update) {
        if(tlActiveSession.get() == null) {
            return updateResult;
        }
        return tlActiveSession.get().load(updateResult, update);
    }

    @Override
    public <T extends ICdmBase> MergeResult<T> load(MergeResult<T> mergeResult, boolean update) {
        if(tlActiveSession.get() == null) {
            return mergeResult;
        }
        return tlActiveSession.get().load(mergeResult, update);
    }

    @Override
    public <T extends CdmBase> Collection<T> load(Collection<T> cdmBaseList, boolean update) {
        if(tlActiveSession.get() == null) {
            return cdmBaseList;
        }
        return tlActiveSession.get().load(cdmBaseList, update);
    }

    void remove(ICdmEntitySessionEnabled owner) {
        ICdmEntitySession session = ownerSessionMap.get(owner);
        if(session == null) {
            logger.info("No Session connected to owner, nothing to do");
            return;
        }
        if(session == tlActiveSession.get()) {
            setActiveSession(null);
        }
        ownerSessionMap.remove(owner);
        notifyObservers();

    }

    @Override
    public void dispose(ICdmEntitySessionEnabled owner) {
        ICdmEntitySession session = ownerSessionMap.get(owner);
        if(session != null) {
            session.dispose();
        }
        if(nullSession != null && nullSession.getOwner() == owner) {
            nullSession = null;
        }
    }

    @Override
    public void disposeAll() {
        Set<ICdmEntitySessionEnabled> owners =
                new HashSet<ICdmEntitySessionEnabled>(ownerSessionMap.keySet());
        for(ICdmEntitySessionEnabled owner : owners) {
            ICdmEntitySession session = ownerSessionMap.get(owner);
            if(session != null) {
                session.dispose();
            }
        }
    }

    public void addToOwnerSessionMap(ICdmEntitySessionEnabled owner, ICdmEntitySession session) {
        ownerSessionMap.put(owner, session);
        notifyObservers();
    }

    @Override
    public void addSessionObserver(ICdmEntitySessionManagerObserver sessionObserver) {
        sessionObservers.add(sessionObserver);
    }

    public void notifyObservers() {
        for(ICdmEntitySessionManagerObserver sessionObserver : sessionObservers) {
            sessionObserver.changed();
        }
    }

    @Override
    public boolean isRemoting() {
        // FIXME:Remoting stupid method to check whether we are in remoting
        return true;
    }

}

