/**
* Copyright (C) 2015 EDIT
* European Distributed Institute of Taxonomy
* http://www.e-taxonomy.eu
*
* The contents of this file are subject to the Mozilla Public License Version 1.1
* See LICENSE.TXT at the top of this package for the full license terms.
*/
package eu.etaxonomy.taxeditor.session;

import java.util.Collection;

import eu.etaxonomy.cdm.api.service.UpdateResult;
import eu.etaxonomy.cdm.cache.CdmTransientEntityCacher;
import eu.etaxonomy.cdm.model.common.CdmBase;
import eu.etaxonomy.cdm.model.common.ICdmBase;
import eu.etaxonomy.cdm.persistence.dto.MergeResult;

/**
 * Manager to manage all {@link ICdmEntitySession}s.
 * <p>
 * It allows to create new sessions, handles binding and disposing, registering
 * and notifying observers, thread handling and also offers
 * convenient methods to load the active session.
 *
 * @author c.mathew
 * @date 2014
 */
public interface ICdmEntitySessionManager {

    /**
     * Creates a new session for the sessionOwner.
     *
     * @param sessionOwner the session owner
     * @param setAsActive if
     * @return the new session
     */
    public abstract ICdmEntitySession newSession(ICdmEntitySessionEnabled sessionOwner, boolean setAsActive);

    /**
     * Makes the {@link NullSession} the active session.
     * @return the null session
     */
    public ICdmEntitySession bindNullSession();

	/**
	 * Makes the session belonging to the session owner the active session.
	 *
	 * @param sessionOwner
	 */
	public abstract void bind(ICdmEntitySessionEnabled sessionOwner);

	/**
	 * Loads an object intto the active session and returns the cached object
	 * which is the identical (same) object if an equal object did not yet exist
	 * in the session. If an equal object existed the cached object is returned.
	 * This cached object is equal (x.equals(x')) but not identical (x != x').
	 *
	 * @param object the object to load into the cache
	 * @param update if <code>true</code>, the state of the object graph
	 *   attached to the object is copied on the according objects in the cache
	 * @return the cached object
	 * @see CdmTransientEntityCacher#load(Object, boolean)
	 * @see CdmTransientEntityCacher#load(CdmBase, boolean)
     */
	public abstract <T extends Object> T load(T object, boolean update);

    /**
     * Loads an {@link CdmBase} into the active session as explained in {@link #load(Object, boolean)}.
     *
     * @see #load(Object, boolean)
     * @see CdmTransientEntityCacher#load(CdmBase, boolean)
     */
	public abstract <T extends CdmBase> T load(T cdmBase, boolean update);

    /**
     * Loads an {@link UpdateResult} into the active session as explained in {@link #load(Object, boolean)}.
     *
     * @see #load(Object, boolean)
     * @see CdmTransientEntityCacher#load(UpdateResult, boolean)
     */
	public abstract UpdateResult load(UpdateResult updateResult, boolean update);

    /**
     * Loads an {@link MergeResult} into the active session as explained in {@link #load(Object, boolean)}.
     *
     * @see #load(Object, boolean)
     * @see CdmTransientEntityCacher#load(MergeResult, boolean)
     */
	public abstract <T extends ICdmBase> MergeResult<T> load(MergeResult<T> mergeResult, boolean update);

    /**
     * Loads an {@link Collection} of {@link CdmBase} objects into the active session as
     * explained in {@link #load(Object, boolean)}.
     *
     * @see #load(Object, boolean)
     * @see CdmTransientEntityCacher#load(Collection, boolean)
     */
	public abstract <T extends CdmBase> Collection<T> load(Collection<T> cdmBaseList, boolean update);

    /**
     * @return the active session
     */
    public ICdmEntitySession getActiveSession();

    /**
     * Returns all sessions managed by this {@link ICdmEntitySessionManager session manager}
     * @return
     */
    public Collection<ICdmEntitySession> getSessions();

    public void addSessionObserver(ICdmEntitySessionManagerObserver sessionObserver);

    public boolean isRemoting();

    public void dispose(ICdmEntitySessionEnabled owner);

    public void disposeAll();

    public <T extends CdmBase> void update();

    /**
     * Returns <code>true</code> if there is a session managed for the given session owner.
     * @param sessionOwner the session owner
     * @return <code>true</code> if a session exists
     */
    public boolean contains(ICdmEntitySessionEnabled sessionOwner);

    /**
     * Returns the {@link NullSession} singleton instance (within this session manager)
     * @return the null session
     */
    public ICdmEntitySession getNullSession();
}